import setuptools

setuptools.setup(
    version="2.0.4",
    long_description="To fit geographically weighted model, "
                     "multiscale geographically weighted regression model, "
                     "geographically and temporally weighted regression model and "
                     "multiscale geographically and temporally weighted regression model.",
    author="Kun Sun",
    author_email="849024477@qq.com",
    packages=['mgtwr'],
    url="https://github.com/sunkun1997/mgtwr",
)
